package ru.landgrafhomyak.db.raw_sql_skeleton

public class Namespace(
	public val name: String,
	private val _path: Array<String>
) {
	private val _usedNames = HashSet<String>()

	private fun _allocName(name: String): String {
		if (!this._usedNames.add(name))
			throw IllegalArgumentException("Name duplication: $name")
		return name
	}

	private fun _formatTableName(name: String) = this._path.joinToString(prefix = "::", separator = "::", postfix = "::${this.name}::${name}")

	public fun subnamespace(name: String): Namespace =
		Namespace(this._allocName(name), this._path + this.name)


	public fun table(name: String): Table {
		return Table(this, this._formatTableName(this._allocName(name)))
	}

	public fun <tUE : Any> table(name: String, constructor: TableConstructor<tUE>): tUE {
		val table = this.table(name)
		return constructor.createTable(table, TableConstructor.Scope(table))
	}

	public fun <tUE : Any> table(name: String, constructor: (table: Table, members: TableConstructor.Scope) -> tUE): tUE =
		this.table(name, DefaultTableConstructorWrapper(constructor))

	public val path: List<String> = this._path.asList()

	private class DefaultTableConstructorWrapper<tUE : Any>(private val fn: (table: Table, members: TableConstructor.Scope) -> tUE) : TableConstructor<tUE> {
		override fun createTable(table: Table, members: TableConstructor.Scope): tUE = this.fn(table, members)
	}
}